Symfony Messenger integration

https://www.drupal.org/project/sm

[[_TOC_]]

# License

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

## Installation

### Requirements

 - Drupal 9 or later, and two core patches:
   - [Patch #1][patch-1]
   - [Patch #2][patch-2] (Already included in Drupal 11)

This project requires Composer to manage dependencies. The Drupal community
recommends [cweagans/composer-patches][cweagans-composer-patches] project to
manage and apply patches. See [Manage dependencies][manage-deps].

 [patch-1]: https://www.drupal.org/project/drupal/issues/2961380 "Support IteratorArgument in \Drupal\Component\DependencyInjection\Dumper\OptimizedPhpArrayDumper::dumpValue"
 [patch-2]: https://www.drupal.org/project/drupal/issues/3108020 "Support ServiceClosureArgument in \Drupal\Component\DependencyInjection\Dumper\OptimizedPhpArrayDumper::dumpValue"
 [manage-deps]: https://www.drupal.org/docs/develop/using-composer/manage-dependencies
 [cweagans-composer-patches]: https://github.com/cweagans/composer-patches

## Site builders

### Transports

Messages are processed immediately by default. To send messages such that they
are stored before processing, enable one or more transports. Messages are then
configured to use specific transports with the _senders map_.

#### Transport/sender map

The sender map determines the routing of how messages are mapped to transports.
This routing map is defined, or optionally sm_config submodule.
See Settings -> Container below for examples.

#### Transports

 - [SQL database transport (via Doctrine)][doctrine-transport]
   A full-featured transport utilising the active Drupal database.

### Console application

#### Consume command

A consume command is included with the project.

```sh
# From the site root.
# Depending on Composer `bin` directory location, the command will be either:
./bin/sm messenger:consume BUSNAMES
./vendor/bin/sm messenger:consume BUSNAMES
```

Replace BUSNAMES with one or more names (space separated) of
receivers/transports to consume in order of priority.

For example `doctrine` if utilising the
[doctrine transport][doctrine-transport]: `./bin/sm messenger:consume doctrine`

Consider running with `--memory-limit` and `--limit` options.

Since the command is designed to run indefinitely, consider combining with a
solution such as [Supervisor][supervisor]. If site infrastructure does not
support long-lived commands, configure the command with server-side executed
cron and the `--time-limit` option.

For additional features or help, see
`sm messenger:consume --help`.

 [doctrine-transport]: https://www.drupal.org/project/symfony_messenger_doctrine "Drupal Symfony Messenger Doctrine transport"
 [supervisor]: https://symfony.com/doc/current/messenger.html#supervisor-configuration

### Replace Core Queue with Symfony Messenger

This project is capable of intercepting legacy Drupal queue items and routing
them through the messenger bus. While retaining existing plugin definition and
process logic, and without any configuration or logic changes for each queue
implementation.

Add to `settings.php`:

```php
# Add after module is enabled in Drupal!
$settings['queue_default'] = \Drupal\sm\QueueInterceptor\SmLegacyQueueFactory::class;
```

If there are other `queue_service_` or `queue_reliable_service_` prefixed
entries in `settings.php`, these will continue to override the default queue
factory. Remove these overrides or modify the factory to use the above class.

See `\Drupal\Core\Queue\QueueFactory::get` for more information on the mechanics
on the Drupal queue factory.

**Runners**

Since queue items will be routed into Symfony Messenger, ensure the consume
command is processing transports. Queue items will not be run via Drupal cron,
Web cron, or Drush queue commands. See _Consume command_ for how to run.

### Settings

The module is primarily configured using container parameters or services.

#### Container

The following is an example of possible configuration. Add this in a site-wide
services YAML referenced by `settings.php`.

```yaml
parameters:
  # These values are not flattened into a hierarchy under a top level `sm` key,
  # so they can be overridden individually.
  sm.routing:
    Drupal\my_module\MyMessage: mytransport
    Drupal\my_module\MyMessage2: [mytransport1, mytransport2]
    'Drupal\my_module\*': mytransport
    '*': mytransport

  # Set the default bus.
  # Value is the ID of the auto-generated service.
  sm.default_bus: sm.bus.my_bus

  sm.buses:
    # The key here is transformed into a service with this value as service ID.
    # i.e. `'sm.bus' . $key`
    my_bus:
      # Set middleware IDs. If default_middleware is enabled below, then these
      # middlewares are inserted in between the _before_ and _after_ middleware.
      middleware: []
      # List of default middleware as found in SymfonyMessengerCompilerPass.
      default_middleware:
        enabled: true
```

#### Settings via UI/Drupal Configuration

A submodule is included which allows configuring via UI and persisting
configuration with the Drupal configuration system. This submodule takes the
configuration and modifies container parameters.

Use both YAML and the submodule together: Container configuration from this
submodule is additive (merged) to parameters provided by YAML;

Note: uninstalling the module will also remove configuration, which makes it
different to a traditional _UI_ modules like _Field UI_ or _Views UI_.

## Developers

### Creating a new message and message handler

There are only two requirements for a message. A message class, and a message
handler.

A message can be any object. There are no attributes, interfaces, service
registration, etc, required.

Starter template:

```php
<?php

declare(strict_types = 1);

namespace Drupal\my_module;

final class MyMessage {

  public function __construct(public string $foo) {}

}
```

A message handler is a class residing in the src/Messenger directory of a
module. It has a simple structure, utilising the `#[AsMessenger]` attribute,
and the PHP magic method `__invoke`.

Starter template:

```php
<?php

declare(strict_types = 1);

namespace Drupal\my_module\Messenger;

use Symfony\Component\Messenger\Attribute\AsMessageHandler;

#[AsMessageHandler]
final class MyMessageHandler {

  public function __invoke(\Drupal\my_module\MyMessage $message): void {
    // Do something with $message.
  }

}
```

Replace `my_module` with a module name, MyMessageHandler with the class name,
and MyMessage with the message class name

For dependency injection, these message handlers have autowiring enabled.

There are more advanced ways to define message handlers, such as with
service.yml file, which is useful for explicit dependency injection. Examples
can be found in tests.

Message handler class discovery can be triggered with cache clear (`drush cr`).

### Dispatch messages

Dispatching a message involves instantiation of the message and sending to a bus:

**Default bus**

If unsure which bus, try the default bus.

```php
<?php
/** @var \Symfony\Component\Messenger\MessageBusInterface $bus */
$bus = \Drupal::service('messenger.default_bus');
$message = new \Drupal\my_module\MyMessage(foo: 'bar');
$bus->dispatch($message);
```

Specific bus

```php
// Either:
// - add dependency to the bus (only valid for compiled dependencies):
//   sm.bus.BUSNAME
// - autowire \Symfony\Component\Messenger\RoutableMessageBus and apply the bus
//   stamp.
/** @var \Symfony\Component\Messenger\RoutableMessageBus $buses */
$buses = \Drupal::service(\Symfony\Component\Messenger\RoutableMessageBus::class);
$message = new \Drupal\my_module\MyMessage(foo: 'bar');
$envelope = new Envelope($message, [
  new BusNameStamp('sm.bus.BUSNAME'),
]);
$buses->dispatch($envelope);
```
